"""Public API for General Cross-Chain Adaptations"""

from typing import Callable, NamedTuple, Tuple

import jax

from blackjax.types import PRNGKey, PyTree


class ChainState(NamedTuple):
    """State of all chains in cross-chain adaptation.

    We store the current iteration for adaptive mechanisms that need this information.
    """

    states: NamedTuple
    current_iter: int


def cross_chain(
    kernel_factory: Callable,
    adaptation_fn: Callable,
):
    """Cross-chain adaptation scheme for general adaptation mechanisms.

    Uses a, preferably large, group of chains proposing new values independently
    but using the same kernel with fixed parameters set using a parameter generating
    mechanism that learns new parameters with information from the last iteration
    of each chain grouped together and taken as independent samples from the target
    density. Samples generated during this warm-up phase for each chain create a
    Markov Chain which is NOT ergodic and should NOT be used as working samples.
    Instead, their utility is in training the parameters of the Markov kernel that
    then should be fixed and used to generate working samples.

    Parameters
    ----------
    kernel_factory
        Function that takes as input the parameters that need to be learned and
        outputs a batched kernel that generates new samples for each chain.
    adaptation_fn
        Function that takes as input the last state of all chains stacked on the
        first dimension, the current iteration of the warm-up phase, and other
        optional inputs and outputs updated parameters.

    Returns
    -------
    init
        Function that initializes the warmup.
    update
        Function that moves the warmup one step.
    """

    def init(initial_states: NamedTuple) -> ChainState:
        return ChainState(initial_states, 0)

    def update(
        rng_key: PRNGKey, state: ChainState, *param
    ) -> Tuple[ChainState, PyTree, NamedTuple]:
        parameters = adaptation_fn(state.states, state.current_iter, *param)
        kernel = kernel_factory(*parameters)
        num_chain = jax.tree_util.tree_leaves(
            jax.tree_map(lambda s: s.shape[0], state.states)
        )[0]
        keys = jax.random.split(rng_key, num_chain)
        new_states, infos = kernel(keys, state.states)
        return ChainState(new_states, state.current_iter + 1), parameters, infos

    return init, update



"""Public API for Maximum-Eigenvalue Adaptation of Damping and Step-size kernel"""

from typing import Callable

import jax
import jax.numpy as jnp

import blackjax.adaptation.chain_adaptation as chain_adaptation
from blackjax.types import PyTree


def base(
    kernel_factory: Callable,
    logprob_grad_fn: Callable,
    batch_fn: Callable = jax.vmap,
):
    """Maximum-Eigenvalue Adaptation of Damping and Step size for the Generalized
    Hamiltonian Monte Carlo kernel [1]_.

    Performs a cross-chain adaptation scheme for the Generalized HMC algorithm that
    automatically selects values for the Generalized HMC's tunable parameters
    based on statistics collected from a population of many chains. It uses heuristics
    determined by the maximum eigenvalue of the covariance and gradient matrices given
    by the grouped samples of all chains with shape (num_chain, parameter dimension).
    Specifically, it does Algorithm 3 of [1]_ with cross-chain adaptation instead of
    parallel ensample chain adaptation.

    Parameters
    ----------
    kernel_factory
        Function that takes as input the step size, alpha and delta parameters
        and outputs a Generalized HMC kernel that generates new samples.
    logprob_grad_fn
        The gradient of logprob_fn, outputs the gradient PyTree for sample.
    batch_fn
        Either jax.vmap or jax.pmap to perform parallel operations.

    Returns
    -------
    init
        Function that initializes the warmup.
    update
        Function that moves the warmup one step.
    final
        Function that returns step size, alpha and delta given a cross-chain
        warmup state.

    References
    ----------
    .. [1]: Hoffman, M. D., & Sountsov, P. (2022). Tuning-Free Generalized
            Hamiltonian Monte Carlo. In International Conference on Artificial
            Intelligence and Statistics (pp. 7799-7813). PML
    """

    def maximum_eigenvalue(matrix: PyTree):
        """Estimate largest eigenvalues of a matrix.

        Calculates an unbiased estimate of the ratio between the sum of the squared
        eigenvalues and the sum of the eigenvalues from the input matrix. This ratio
        approximates the largest eigenvalue well except in cases when there are a
        large number of small eigenvalues significantly larger than 0 but significantly
        smaller than the largest eigenvalue. This unbiased estimate is used instead of
        directly computing an unbiased estimate of the largest eigenvalue because of
        the latter's large variance.
        """

        X = jnp.vstack(
            [
                leaf.reshape(leaf.shape[0], -1).T
                for leaf in jax.tree_util.tree_leaves(matrix)
            ]
        ).T
        n, _ = X.shape
        S = X @ X.T
        diag_S = jnp.diag(S)
        lamda = jnp.sum(diag_S) / n
        lamda_sq = (jnp.sum(S**2) - jnp.sum(diag_S**2)) / (n * (n - 1))
        return lamda_sq / lamda

    def parameter_gn(batch_state, current_iter):
        """Generate updated parameteters using information of cross-chain samples.

        Using heuristics from the maximum eigenvalue of the covariance and gradient
        matrices given by the grouped samples of all chains, find new optimal
        parameters for the Generalized HMC kernel.
        """

        batch_position = batch_state.position
        mean_position = jax.tree_map(lambda p: p.mean(axis=0), batch_position)
        sd_position = jax.tree_map(lambda p: p.std(axis=0), batch_position)
        batch_norm = jax.tree_map(
            lambda p, mu, sd: (p - mu) / sd,
            batch_position,
            mean_position,
            sd_position,
        )
        batch_grad = batch_fn(logprob_grad_fn)(batch_position)
        batch_grad_scaled = jax.tree_map(
            lambda grad, sd: grad * sd, batch_grad, sd_position
        )
        epsilon = jnp.minimum(
            0.5 / jnp.sqrt(maximum_eigenvalue(batch_grad_scaled)), 1.0
        )
        gamma = jnp.maximum(
            1.0 / jnp.sqrt(maximum_eigenvalue(batch_norm)),
            1.0 / ((current_iter + 1) * epsilon),
        )
        alpha = 1.0 - jnp.exp(-2.0 * epsilon * gamma)
        delta = alpha / 2
        step_size = jax.tree_map(lambda sd: epsilon * sd, sd_position)
        return step_size, alpha, delta

    init, update = chain_adaptation.cross_chain(
        lambda *parameters: batch_fn(kernel_factory(*parameters)),
        parameter_gn,
    )

    def final(last_state: chain_adaptation.ChainState) -> PyTree:
        parameters = parameter_gn(
            last_state.states,
            last_state.current_iter,
        )
        return kernel_factory(*parameters)

    return init, update, final



from typing import Callable, Dict, Optional, Union

import jax
import jax.numpy as jnp

import blackjax.adaptation as adaptation
import blackjax.mcmc as mcmc
import blackjax.sgmcmc as sgmcmc
import blackjax.smc as smc
import blackjax.vi as vi
from blackjax.base import AdaptationAlgorithm, SamplingAlgorithm
from blackjax.progress_bar import progress_bar_scan
from blackjax.types import Array, PRNGKey, PyTree

__all__ = [
    "adaptive_tempered_smc",
    "hmc",
    "mala",
    "nuts",
    "orbital_hmc",
    "rmh",
    "sgld",
    "sghmc",
    "tempered_smc",
    "window_adaptation",
    "irmh",
    "pathfinder",
    "pathfinder_adaptation",
    "mgrad_gaussian",
]


# -----------------------------------------------------------------------------
#                           SEQUENTIAL MONTE CARLO
# -----------------------------------------------------------------------------


class adaptive_tempered_smc:
    """Implements the (basic) user interface for the Adaptive Tempered SMC kernel.

    Returns
    -------
    A ``SamplingAlgorithm``.

    """

    init = staticmethod(smc.tempered.init)
    kernel = staticmethod(smc.adaptive_tempered.kernel)

    def __new__(  # type: ignore[misc]
        cls,
        logprior_fn: Callable,
        loglikelihood_fn: Callable,
        mcmc_algorithm: SamplingAlgorithm,
        mcmc_parameters: Dict,
        resampling_fn: Callable,
        target_ess: float,
        root_solver: Callable = smc.solver.dichotomy,
        use_log_ess: bool = True,
        mcmc_iter: int = 10,
    ) -> SamplingAlgorithm:
        def kernel_factory(logprob_fn):
            return mcmc_algorithm(logprob_fn, **mcmc_parameters).step

        step = cls.kernel(
            logprior_fn,
            loglikelihood_fn,
            kernel_factory,
            mcmc_algorithm.init,
            resampling_fn,
            target_ess,
            root_solver,
            use_log_ess,
            mcmc_iter,
        )

        def init_fn(position: PyTree):
            return cls.init(position)

        def step_fn(rng_key: PRNGKey, state):
            return step(
                rng_key,
                state,
            )

        return SamplingAlgorithm(init_fn, step_fn)


class tempered_smc:
    """Implements the (basic) user interface for the Adaptive Tempered SMC kernel.


    Returns
    -------
    A ``SamplingAlgorithm``.

    """

    init = staticmethod(smc.tempered.init)
    kernel = staticmethod(smc.tempered.kernel)

    def __new__(  # type: ignore[misc]
        cls,
        logprior_fn: Callable,
        loglikelihood_fn: Callable,
        mcmc_algorithm: SamplingAlgorithm,
        mcmc_parameters: Dict,
        resampling_fn: Callable,
        mcmc_iter: int = 10,
    ) -> SamplingAlgorithm:
        def kernel_factory(logprob_fn):
            return mcmc_algorithm(logprob_fn, **mcmc_parameters).step

        step = cls.kernel(
            logprior_fn,
            loglikelihood_fn,
            kernel_factory,
            mcmc_algorithm.init,
            resampling_fn,
            mcmc_iter,
        )

        def init_fn(position: PyTree):
            return cls.init(position)

        def step_fn(rng_key: PRNGKey, state, lmbda):
            return step(
                rng_key,
                state,
                lmbda,
            )

        return SamplingAlgorithm(init_fn, step_fn)  # type: ignore[arg-type]


# -----------------------------------------------------------------------------
#                         MARKOV CHAIN MONTE CARLO
# -----------------------------------------------------------------------------


class hmc:
    """Implements the (basic) user interface for the HMC kernel.

    The general hmc kernel (:meth:`blackjax.mcmc.hmc.kernel`, alias `blackjax.hmc.kernel`) can be
    cumbersome to manipulate. Since most users only need to specify the kernel
    parameters at initialization time, we provide a helper function that
    specializes the general kernel.

    We also add the general kernel and state generator as an attribute to this class so
    users only need to pass `blackjax.hmc` to SMC, adaptation, etc. algorithms.

    Examples
    --------

    A new HMC kernel can be initialized and used with the following code:

    .. code::

        hmc = blackjax.hmc(logprob_fn step_size, inverse_mass_matrix, num_integration_steps)
        state = hmc.init(position)
        new_state, info = hmc.step(rng_key, state)

    Kernels are not jit-compiled by default so you will need to do it manually:

    .. code::

       step = jax.jit(hmc.step)
       new_state, info = step(rng_key, state)

    Should you need to you can always use the base kernel directly:

    .. code::

       import blackjax.mcmc.integrators as integrators

       kernel = blackjax.hmc.kernel(integrators.mclachlan)
       state = blackjax.hmc.init(position, logprob_fn)
       state, info = kernel(rng_key, state, logprob_fn, step_size, inverse_mass_matrix, num_integration_steps)

    Parameters
    ----------
    logprob_fn
        The logprobability density function we wish to draw samples from. This
        is minus the potential function.
    step_size
        The value to use for the step size in the symplectic integrator.
    inverse_mass_matrix
        The value to use for the inverse mass matrix when drawing a value for
        the momentum and computing the kinetic energy.
    num_integration_steps
        The number of steps we take with the symplectic integrator at each
        sample step before returning a sample.
    divergence_threshold
        The absolute value of the difference in energy between two states above
        which we say that the transition is divergent. The default value is
        commonly found in other libraries, and yet is arbitrary.
    integrator
        (algorithm parameter) The symplectic integrator to use to integrate the trajectory.\

    Returns
    -------
    A ``SamplingAlgorithm``.
    """

    init = staticmethod(mcmc.hmc.init)
    kernel = staticmethod(mcmc.hmc.kernel)

    def __new__(  # type: ignore[misc]
        cls,
        logprob_fn: Callable,
        step_size: float,
        inverse_mass_matrix: Array,
        num_integration_steps: int,
        *,
        divergence_threshold: int = 1000,
        integrator: Callable = mcmc.integrators.velocity_verlet,
        logprob_grad_fn: Optional[Callable] = None,
    ) -> SamplingAlgorithm:
        step = cls.kernel(integrator, divergence_threshold)

        def init_fn(position: PyTree):
            return cls.init(position, logprob_fn, logprob_grad_fn)

        def step_fn(rng_key: PRNGKey, state):
            return step(
                rng_key,
                state,
                logprob_fn,
                step_size,
                inverse_mass_matrix,
                num_integration_steps,
                logprob_grad_fn=logprob_grad_fn,
            )

        return SamplingAlgorithm(init_fn, step_fn)


class mala:
    """Implements the (basic) user interface for the MALA kernel.

    The general mala kernel (:meth:`blackjax.mcmc.mala.kernel`, alias `blackjax.mala.kernel`) can be
    cumbersome to manipulate. Since most users only need to specify the kernel
    parameters at initialization time, we provide a helper function that
    specializes the general kernel.

    We also add the general kernel and state generator as an attribute to this class so
    users only need to pass `blackjax.mala` to SMC, adaptation, etc. algorithms.

    Examples
    --------

    A new MALA kernel can be initialized and used with the following code:

    .. code::

        mala = blackjax.mala(logprob_fn, step_size)
        state = mala.init(position)
        new_state, info = mala.step(rng_key, state)

    Kernels are not jit-compiled by default so you will need to do it manually:

    .. code::

       step = jax.jit(mala.step)
       new_state, info = step(rng_key, state)

    Should you need to you can always use the base kernel directly:

    .. code::

       kernel = blackjax.mala.kernel(logprob_fn)
       state = blackjax.mala.init(position, logprob_fn)
       state, info = kernel(rng_key, state, logprob_fn, step_size)

    Parameters
    ----------
    logprob_fn
        The logprobability density function we wish to draw samples from. This
        is minus the potential function.
    step_size
        The value to use for the step size in the symplectic integrator.

    Returns
    -------
    A ``SamplingAlgorithm``.

    """

    init = staticmethod(mcmc.mala.init)
    kernel = staticmethod(mcmc.mala.kernel)

    def __new__(  # type: ignore[misc]
        cls,
        logprob_fn: Callable,
        step_size: float,
    ) -> SamplingAlgorithm:
        step = cls.kernel()

        def init_fn(position: PyTree):
            return cls.init(position, logprob_fn)

        def step_fn(rng_key: PRNGKey, state):
            return step(rng_key, state, logprob_fn, step_size)

        return SamplingAlgorithm(init_fn, step_fn)


class nuts:
    """Implements the (basic) user interface for the nuts kernel.

    Examples
    --------

    A new NUTS kernel can be initialized and used with the following code:

    .. code::

        nuts = blackjax.nuts(logprob_fn step_size, inverse_mass_matrix)
        state = nuts.init(position)
        new_state, info = nuts.step(rng_key, state)

    We can JIT-compile the step function for more speed:

    .. code::

        step = jax.jit(nuts.step)
        new_state, info = step(rng_key, state)

    You can always use the base kernel should you need to:

    .. code::

       import blackjax.mcmc.integrators as integrators

       kernel = blackjax.nuts.kernel(integrators.yoshida)
       state = blackjax.nuts.init(position, logprob_fn)
       state, info = kernel(rng_key, state, logprob_fn, step_size, inverse_mass_matrix)

    Parameters
    ----------
    logprob_fn
        The logprobability density function we wish to draw samples from. This
        is minus the potential function.
    step_size
        The value to use for the step size in the symplectic integrator.
    inverse_mass_matrix
        The value to use for the inverse mass matrix when drawing a value for
        the momentum and computing the kinetic energy.
    max_num_doublings
        The maximum number of times we double the length of the trajectory before
        returning if no U-turn has been obserbed or no divergence has occured.
    divergence_threshold
        The absolute value of the difference in energy between two states above
        which we say that the transition is divergent. The default value is
        commonly found in other libraries, and yet is arbitrary.
    integrator
        (algorithm parameter) The symplectic integrator to use to integrate the trajectory.

    Returns
    -------
    A ``SamplingAlgorithm``.

    """

    init = staticmethod(mcmc.hmc.init)
    kernel = staticmethod(mcmc.nuts.kernel)

    def __new__(  # type: ignore[misc]
        cls,
        logprob_fn: Callable,
        step_size: float,
        inverse_mass_matrix: Array,
        *,
        max_num_doublings: int = 10,
        divergence_threshold: int = 1000,
        integrator: Callable = mcmc.integrators.velocity_verlet,
        logprob_grad_fn: Optional[Callable] = None,
    ) -> SamplingAlgorithm:
        step = cls.kernel(integrator, divergence_threshold, max_num_doublings)

        def init_fn(position: PyTree):
            return cls.init(position, logprob_fn)

        def step_fn(rng_key: PRNGKey, state):
            return step(
                rng_key,
                state,
                logprob_fn,
                step_size,
                inverse_mass_matrix,
                logprob_grad_fn=logprob_grad_fn,
            )

        return SamplingAlgorithm(init_fn, step_fn)


class mgrad_gaussian:
    """Implements the marginal sampler for latent Gaussian model of [1].

    It uses a first order approximation to the log_likelihood of a model with Gaussian prior.
    Interestingly, the only parameter that needs calibrating is the "step size" delta, which can be done very efficiently.
    Calibrating it to have an acceptance rate of roughly 50% is a good starting point.

    Examples
    --------
    A new marginal latent Gaussian MCMC kernel for a model q(x) ∝ exp(f(x)) N(x; m, C) can be initialized and
    used for a given "step size" delta with the following code:
    .. code::

        mgrad_gaussian = blackjax.mgrad_gaussian(f, C, use_inverse=False, mean=m)
        state = latent_gaussian.init(zeros)  # Starting at the mean of the prior
        new_state, info = mgrad_gaussian.step(rng_key, state, delta)

    We can JIT-compile the step function for better performance

    .. code::
        step = jax.jit(latent_gaussian.step)
        new_state, info = step(rng_key, state, delta)

    Parameters
    ----------
    logprob_fn
        The logarithm of the likelihood function for the latent Gaussian model.
    covariance
        The covariance of the prior Gaussian density.
    mean: optional
        Mean of the prior Gaussian density. Default is zero.

    Returns
    -------
    A ``SamplingAlgorithm``.

    References
    ----------
    [1]: Titsias, M.K. and Papaspiliopoulos, O. (2018), Auxiliary gradient-based sampling algorithms. J. R. Stat. Soc. B, 80: 749-767. https://doi.org/10.1111/rssb.12269
    """

    def __new__(  # type: ignore[misc]
        cls,
        logprob_fn: Callable,
        covariance: Array,
        mean: Optional[Array] = None,
    ) -> SamplingAlgorithm:
        init, step = mcmc.marginal_latent_gaussian.init_and_kernel(
            logprob_fn, covariance, mean
        )

        def init_fn(position: Array):
            return init(position)

        def step_fn(rng_key: PRNGKey, state, delta: float):
            return step(
                rng_key,
                state,
                delta,
            )

        return SamplingAlgorithm(init_fn, step_fn)  # type: ignore[arg-type]


# -----------------------------------------------------------------------------
#                        STOCHASTIC GRADIENT MCMC
# -----------------------------------------------------------------------------


class sgld:
    """Implements the (basic) user interface for the SGLD kernel.

    The general sgld kernel (:meth:`blackjax.mcmc.sgld.kernel`, alias `blackjax.sgld.kernel`) can be
    cumbersome to manipulate. Since most users only need to specify the kernel
    parameters at initialization time, we provide a helper function that
    specializes the general kernel.

    Example
    -------

    To initialize a SGLD kernel one needs to specify a schedule function, which
    returns a step size at each sampling step, and a gradient estimator
    function. Here for a constant step size, and `data_size` data samples:

    .. code::

        schedule_fn = lambda _: 1e-3
        grad_fn = blackjax.sgmcmc.gradients.grad_estimator(logprior_fn, loglikelihood_fn, data_size)

    We can now initialize the sgld kernel and the state:

    .. code::

        sgld = blackjax.sgld(grad_fn, schedule_fn)
        state = sgld.init(position)

    Assuming we have an iterator `batches` that yields batches of data we can perform one step:

    .. code::

        data_batch = next(batches)
        new_state = sgld.step(rng_key, state, data_batch)

    Kernels are not jit-compiled by default so you will need to do it manually:

    .. code::

       step = jax.jit(sgld.step)
       new_state, info = step(rng_key, state)

    Parameters
    ----------
    gradient_estimator_fn
       A function which, given a position and a batch of data, returns an estimation
       of the value of the gradient of the log-posterior distribution at this position.
    schedule_fn
       A function which returns a step size given a step number.

    Returns
    -------
    A ``SamplingAlgorithm``.

    """

    init = staticmethod(sgmcmc.sgld.init)
    kernel = staticmethod(sgmcmc.sgld.kernel)

    def __new__(  # type: ignore[misc]
        cls,
        grad_estimator_fn: Callable,
        learning_rate: Union[Callable[[int], float], float],
    ) -> SamplingAlgorithm:

        step = cls.kernel(grad_estimator_fn)

        if callable(learning_rate):
            learning_rate_fn = learning_rate
        elif float(learning_rate):

            def learning_rate_fn(_):
                return learning_rate

        else:
            raise TypeError(
                "The learning rate must either be a float (which corresponds to a constant learning rate) "
                f"or a function of the index of the current iteration. Got {type(learning_rate)} instead."
            )

        def init_fn(position: PyTree, data_batch: PyTree):
            return cls.init(position, data_batch, grad_estimator_fn)

        def step_fn(rng_key: PRNGKey, state, data_batch: PyTree):
            step_size = learning_rate_fn(state.step)
            return step(rng_key, state, data_batch, step_size)

        return SamplingAlgorithm(init_fn, step_fn)  # type: ignore[arg-type]


class sghmc:
    """Implements the (basic) user interface for the SGHMC kernel.

    The general sghmc kernel (:meth:`blackjax.mcmc.sghmc.kernel`, alias `blackjax.sghmc.kernel`) can be
    cumbersome to manipulate. Since most users only need to specify the kernel
    parameters at initialization time, we provide a helper function that
    specializes the general kernel.

    Example
    -------

    To initialize a SGHMC kernel one needs to specify a schedule function, which
    returns a step size at each sampling step, and a gradient estimator
    function. Here for a constant step size, and `data_size` data samples:

    .. code::

        schedule_fn = lambda _: 1e-3
        grad_fn = blackjax.sgmcmc.gradients.grad_estimator(logprior_fn, loglikelihood_fn, data_size)

    We can now initialize the sghmc kernel and the state. Like HMC, SGHMC needs the user to specify a number of integration steps.

    .. code::

        sghmc = blackjax.sghmc(grad_fn, schedule_fn, num_integration_steps)
        state = sghmc.init(position)

    Assuming we have an iterator `batches` that yields batches of data we can perform one step:

    .. code::

        data_batch = next(batches)
        new_state = sghmc.step(rng_key, state, data_batch)

    Kernels are not jit-compiled by default so you will need to do it manually:

    .. code::

       step = jax.jit(sghmc.step)
       new_state, info = step(rng_key, state)

    Parameters
    ----------
    gradient_estimator_fn
       A function which, given a position and a batch of data, returns an estimation
       of the value of the gradient of the log-posterior distribution at this position.
    schedule_fn
       A function which returns a step size given a step number.

    Returns
    -------
    A ``SamplingAlgorithm``.

    """

    init = staticmethod(sgmcmc.sgld.init)
    kernel = staticmethod(sgmcmc.sghmc.kernel)

    def __new__(  # type: ignore[misc]
        cls,
        grad_estimator_fn: Callable,
        learning_rate: Union[Callable[[int], float], float],
        num_integration_steps: int = 10,
    ) -> SamplingAlgorithm:

        step = cls.kernel(grad_estimator_fn)

        if callable(learning_rate):
            learning_rate_fn = learning_rate
        elif float(learning_rate):

            def learning_rate_fn(_):
                return learning_rate

        else:
            raise TypeError(
                "The learning rate must either be a float (which corresponds to a constant learning rate) "
                f"or a function of the index of the current iteration. Got {type(learning_rate)} instead."
            )

        def init_fn(position: PyTree, data_batch: PyTree):
            return cls.init(position, data_batch, grad_estimator_fn)

        def step_fn(rng_key: PRNGKey, state, data_batch: PyTree):
            step_size = learning_rate_fn(state.step)
            return step(rng_key, state, data_batch, step_size, num_integration_steps)

        return SamplingAlgorithm(init_fn, step_fn)  # type: ignore[arg-type]


# -----------------------------------------------------------------------------
#                                 ADAPTATION
# -----------------------------------------------------------------------------


def window_adaptation(
    algorithm: Union[hmc, nuts],
    logprob_fn: Callable,
    num_steps: int = 1000,
    is_mass_matrix_diagonal: bool = True,
    initial_step_size: float = 1.0,
    target_acceptance_rate: float = 0.80,
    progress_bar: bool = False,
    logprob_grad_fn: Optional[Callable] = None,
    **parameters,
) -> AdaptationAlgorithm:
    """Adapt the parameters of algorithms in the HMC family.

    Algorithms in the HMC family on a euclidean manifold depend on the value of
    at least two parameters: the step size, related to the trajectory
    integrator, and the mass matrix, linked to the euclidean metric.

    Good tuning is very important, especially for algorithms like NUTS which can
    be extremely inefficient with the wrong parameter values. This function
    provides a general-purpose algorithm to tune the values of these parameters.
    Originally based on Stan's window adaptation, the algorithm has evolved to
    improve performance and quality.

    Parameters
    ----------
    algorithm
        The algorithm whose parameters are being tuned.
    logprob_fn
        The log density probability density function from which we wish to sample.
    num_steps
        The number of adaptation steps.
    is_mass_matrix_diagonal
        Whether we should adapt a diagonal mass matrix.
    initial_step_size
        The initial step size used in the algorithm.
    target_acceptance_rate
        The acceptance rate that we target during step size adaptation.
    progress_bar
        Whether we should display a progress bar.
    logprob_grad_fn
        The gradient of logprob_fn.  If it's not provided, it will be computed
        by jax using reverse mode autodiff (jax.grad).
    **parameters
        The extra parameters to pass to the algorithm, e.g. the number of
        integration steps for HMC.

    Returns
    -------
    A function that returns the last chain state and a sampling kernel with the tuned parameter values from an initial state.

    """

    kernel = algorithm.kernel()

    def kernel_factory(step_size: float, inverse_mass_matrix: Array):
        def kernel_fn(rng_key, state):
            return kernel(
                rng_key,
                state,
                logprob_fn,
                step_size,
                inverse_mass_matrix,
                logprob_grad_fn=logprob_grad_fn,
                **parameters,
            )

        return kernel_fn

    schedule = adaptation.window_adaptation.schedule(num_steps)
    init, update, final = adaptation.window_adaptation.base(
        kernel_factory,
        is_mass_matrix_diagonal,
        target_acceptance_rate=target_acceptance_rate,
    )

    def one_step(carry, xs):
        _, rng_key, adaptation_stage = xs
        state, adaptation_state = carry
        state, adaptation_state, info = update(
            rng_key, state, adaptation_state, adaptation_stage
        )
        return ((state, adaptation_state), (state, info, adaptation_state))

    def run(rng_key: PRNGKey, position: PyTree):
        init_state = algorithm.init(position, logprob_fn, logprob_grad_fn)
        init_warmup_state = init(init_state, initial_step_size)

        if progress_bar:
            print("Running window adaptation")
            one_step_ = jax.jit(progress_bar_scan(num_steps)(one_step))
        else:
            one_step_ = jax.jit(one_step)

        keys = jax.random.split(rng_key, num_steps)
        last_state, warmup_chain = jax.lax.scan(
            one_step_,
            (init_state, init_warmup_state),
            (jnp.arange(num_steps), keys, schedule),
        )
        last_chain_state, last_warmup_state = last_state

        step_size, inverse_mass_matrix = final(last_warmup_state)
        kernel = kernel_factory(step_size, inverse_mass_matrix)

        return last_chain_state, kernel, warmup_chain

    return AdaptationAlgorithm(run)


def meads(
    logprob_fn: Callable,
    num_chain: int,
    num_steps: int = 1000,
    *,
    divergence_threshold: int = 1000,
    logprob_grad_fn: Optional[Callable] = None,
    batch_fn: Callable = jax.vmap,
) -> AdaptationAlgorithm:
    """Adapt the parameters of the Generalized HMC algorithm.

    The Generalized HMC algorithm depends on three parameters, each controlling
    one element of its behaviour: step size controls the integrator's dynamics,
    alpha controls the persistency of the momentum variable, and delta controls
    the deterministic transformation of the slice variable used to perform the
    non-reversible Metropolis-Hastings accept/reject step.

    The step size parameter is chosen to ensure the stability of the velocity
    verlet integrator, the alpha parameter to make the influence of the current
    state on future states of the momentum variable to decay exponentially, and
    the delta parameter to maximize the acceptance of proposal but with good
    mixing properties for the slice variable. These characteristics are targeted
    by controlling heuristics based on the maximum eigenvalues of the correlation
    and gradient matrices of the cross-chain samples, under simpifyng assumptions.

    Good tuning is fundamental for the non-reversible Generalized HMC sampling
    algorithm to explore the target space efficienty and output uncorrelated, or
    as uncorrelated as possible, samples from the target space. Furthermore, the
    single integrator step of the algorithm lends itself for fast sampling
    on parallel computer architectures.

    Parameters
    ----------
    logprob_fn
        The log density probability density function from which we wish to sample.
    num_chain
        Number of chains used for cross-chain warm-up training.
    num_steps
        The number of adaptation steps.
    divergence_threshold
        Value of the difference in energy above which we consider that the
        transition is divergent.
    logprob_grad_fn
        The gradient of logprob_fn.  If it's not provided, it will be computed
        by jax using reverse mode autodiff (jax.grad).
    batch_fn
        Either jax.vmap or jax.pmap to perform parallel operations.

    Returns
    -------
    A function that returns the last cross-chain state, a sampling kernel with the
    tuned parameter values, and all the warm-up states for diagnostics.

    """

    kernel = ghmc.kernel(divergence_threshold=divergence_threshold)

    def kernel_factory(step_size: PyTree, alpha: float, delta: float):
        def kernel_fn(rng_key, state):
            return kernel(
                rng_key,
                state,
                logprob_fn,
                step_size,
                alpha,
                delta,
                logprob_grad_fn=logprob_grad_fn,
            )

        return kernel_fn

    init, update, final = adaptation.meads.base(
        kernel_factory,
        logprob_grad_fn or jax.grad(logprob_fn),
    )

    batch_init = batch_fn(lambda r, p: ghmc.init(r, p, logprob_fn, logprob_grad_fn))

    def one_step(state, rng_key):
        state, parameters, infos = update(rng_key, state)
        return state, (state, parameters, infos)

    def run(rng_key: PRNGKey, positions: PyTree):

        key_init, key_warm = jax.random.split(rng_key)
        rng_keys = jax.random.split(key_init, num_chain)
        states = batch_init(rng_keys, positions)
        init_state = init(states)

        keys = jax.random.split(key_warm, num_steps)
        last_state, (warmup_states, parameters, info) = jax.lax.scan(
            one_step, init_state, keys
        )
        kernel = final(last_state)

        return last_state, kernel, warmup_states

    return AdaptationAlgorithm(run)  # type: ignore[arg-type]


class rmh:
    """Implements the (basic) user interface for the gaussian random walk kernel

    Examples
    --------

    A new Gaussian Random Walk kernel can be initialized and used with the following code:

    .. code::

        rmh = blackjax.rmh(logprob_fn, sigma)
        state = rmh.init(position)
        new_state, info = rmh.step(rng_key, state)

    We can JIT-compile the step function for better performance

    .. code::

        step = jax.jit(rmh.step)
        new_state, info = step(rng_key, state)

    Parameters
    ----------
    logprob_fn
        The log density probability density function from which we wish to sample.
    sigma
        The value of the covariance matrix of the gaussian proposal distribution.

    Returns
    -------
    A ``SamplingAlgorithm``.

    """

    init = staticmethod(mcmc.rmh.init)
    kernel = staticmethod(mcmc.rmh.kernel)

    def __new__(  # type: ignore[misc]
        cls,
        logprob_fn: Callable,
        sigma: Array,
    ) -> SamplingAlgorithm:
        step = cls.kernel()

        def init_fn(position: PyTree):
            return cls.init(position, logprob_fn)

        def step_fn(rng_key: PRNGKey, state):
            return step(
                rng_key,
                state,
                logprob_fn,
                sigma,
            )

        return SamplingAlgorithm(init_fn, step_fn)


class irmh:
    """Implements the (basic) user interface for the independent RMH.

    Examples
    --------

    A new kernel can be initialized and used with the following code:

    .. code::

        rmh = blackjax.irmh(logprob_fn, proposal_distribution)
        state = rmh.init(position)
        new_state, info = rmh.step(rng_key, state)

    We can JIT-compile the step function for better performance

    .. code::

        step = jax.jit(rmh.step)
        new_state, info = step(rng_key, state)

    Parameters
    ----------
    logprob_fn
        The log density probability density function from which we wish to sample.
    proposal_distribution
        A Callable that takes a random number generator and produces a new proposal. The
        proposal is independent of the sampler's current state.

    Returns
    -------
    A ``SamplingAlgorithm``.

    """

    init = staticmethod(mcmc.rmh.init)
    kernel = staticmethod(mcmc.irmh.kernel)

    def __new__(  # type: ignore[misc]
        cls,
        logprob_fn: Callable,
        proposal_distribution: Callable,
    ) -> SamplingAlgorithm:

        step = cls.kernel(proposal_distribution)

        def init_fn(position: PyTree):
            return cls.init(position, logprob_fn)

        def step_fn(rng_key: PRNGKey, state):
            return step(rng_key, state, logprob_fn)

        return SamplingAlgorithm(init_fn, step_fn)


class orbital_hmc:
    """Implements the (basic) user interface for the Periodic orbital MCMC kernel.

    Each iteration of the periodic orbital MCMC outputs ``period`` weighted samples from
    a single Hamiltonian orbit connecting the previous sample and momentum (latent) variable
    with precision matrix ``inverse_mass_matrix``, evaluated using the ``bijection`` as an
    integrator with discretization parameter ``step_size``.

    Examples
    --------

    A new Periodic orbital MCMC kernel can be initialized and used with the following code:

    .. code::

        per_orbit = blackjax.orbital_hmc(logprob_fn, step_size, inverse_mass_matrix, period)
        state = per_orbit.init(position)
        new_state, info = per_orbit.step(rng_key, state)

    We can JIT-compile the step function for better performance

    .. code::

        step = jax.jit(per_orbit.step)
        new_state, info = step(rng_key, state)

    Parameters
    ----------
    logprob_fn
        The logarithm of the probability density function we wish to draw samples from. This
        is minus the potential energy function.
    step_size
        The value to use for the step size in for the symplectic integrator to buid the orbit.
    inverse_mass_matrix
        The value to use for the inverse mass matrix when drawing a value for
        the momentum and computing the kinetic energy.
    period
        The number of steps used to build the orbit.
    bijection
        (algorithm parameter) The symplectic integrator to use to build the orbit.

    Returns
    -------
    A ``SamplingAlgorithm``.
    """

    init = staticmethod(mcmc.periodic_orbital.init)
    kernel = staticmethod(mcmc.periodic_orbital.kernel)

    def __new__(  # type: ignore[misc]
        cls,
        logprob_fn: Callable,
        step_size: float,
        inverse_mass_matrix: Array,  # assume momentum is always Gaussian
        period: int,
        *,
        bijection: Callable = mcmc.integrators.velocity_verlet,
    ) -> SamplingAlgorithm:
        step = cls.kernel(bijection)

        def init_fn(position: PyTree):
            return cls.init(position, logprob_fn, period)

        def step_fn(rng_key: PRNGKey, state):
            return step(
                rng_key,
                state,
                logprob_fn,
                step_size,
                inverse_mass_matrix,
                period,
            )

        return SamplingAlgorithm(init_fn, step_fn)


class elliptical_slice:
    """Implements the (basic) user interface for the Elliptical Slice sampling kernel.

    Examples
    --------

    A new Elliptical Slice sampling kernel can be initialized and used with the following code:

    .. code::

        ellip_slice = blackjax.elliptical_slice(loglikelihood_fn, cov_matrix)
        state = ellip_slice.init(position)
        new_state, info = ellip_slice.step(rng_key, state)

    We can JIT-compile the step function for better performance

    .. code::

        step = jax.jit(ellip_slice.step)
        new_state, info = step(rng_key, state)

    Parameters
    ----------
    loglikelihood_fn
        Only the log likelihood function from the posterior distributon we wish to sample.
    cov_matrix
        The value of the covariance matrix of the gaussian prior distribution from the posterior we wish to sample.

    Returns
    -------
    A ``SamplingAlgorithm``.
    """

    init = staticmethod(mcmc.elliptical_slice.init)
    kernel = staticmethod(mcmc.elliptical_slice.kernel)

    def __new__(  # type: ignore[misc]
        cls,
        loglikelihood_fn: Callable,
        *,
        mean: Array,
        cov: Array,
    ) -> SamplingAlgorithm:
        step = cls.kernel(cov, mean)

        def init_fn(position: PyTree):
            return cls.init(position, loglikelihood_fn)

        def step_fn(rng_key: PRNGKey, state):
            return step(
                rng_key,
                state,
                loglikelihood_fn,
            )

        return SamplingAlgorithm(init_fn, step_fn)


class ghmc:
    """Implements the (basic) user interface for the Generalized HMC kernel.

    The Generalized HMC kernel performs a similar procedure to the standard HMC
    kernel with the difference of a persistent momentum variable and a non-reversible
    Metropolis-Hastings step instead of the standard Metropolis-Hastings acceptance
    step.

    This means that the sampling of the momentum variable depends on the previous
    momentum, the rate of persistence depends on the alpha parameter, and that the
    Metropolis-Hastings accept/reject step is done through slice sampling with a
    non-reversible slice variable also dependent on the previous slice, the determinisitc
    transformation is defined by the delta parameter.

    The Generalized HMC does not have a trajectory length parameter, it always performs
    one iteration of the velocity verlet integrator with a given step size, making
    the algorithm a good candiate for running many chains in parallel.

    Examples
    --------

    A new Generalized HMC kernel can be initialized and used with the following code:

    .. code::

        ghmc_kernel = blackjax.ghmc(logprob_fn, step_size, alpha, delta)
        state = ghmc_kernel.init(rng_key, position)
        new_state, info = ghmc_kernel.step(rng_key, state)

    We can JIT-compile the step function for better performance

    .. code::

        step = jax.jit(ghmc_kernel.step)
        new_state, info = step(rng_key, state)

    Parameters
    ----------
    logprob_fn
        The logprobability density function we wish to draw samples from. This
        is minus the potential function.
    step_size
        A PyTree of the same structure as the target PyTree (position) with the
        values used for as a step size for each dimension of the target space in
        the velocity verlet integrator.
    alpha
        The value defining the persistence of the momentum variable.
    delta
        The value defining the deterministic translation of the slice variable.
    divergence_threshold
        The absolute value of the difference in energy between two states above
        which we say that the transition is divergent. The default value is
        commonly found in other libraries, and yet is arbitrary.
    noise_gn
        A function that takes as input the slice variable and outputs a random
        variable used as a noise correction of the persistent slice update.
        The parameter defaults to a random variable with a single atom at 0.
    logprob_grad_fn
        Optional function customizing the gradients of the target log density.

    Returns
    -------
    A ``SamplingAlgorithm``.
    """

    init = staticmethod(mcmc.ghmc.init)
    kernel = staticmethod(mcmc.ghmc.kernel)

    def __new__(  # type: ignore[misc]
        cls,
        logprob_fn: Callable,
        step_size: PyTree,
        alpha: float,
        delta: float,
        *,
        divergence_threshold: int = 1000,
        noise_gn: Callable = lambda _: 0.0,
        logprob_grad_fn: Optional[Callable] = None,
    ) -> SamplingAlgorithm:

        step = cls.kernel(noise_gn, divergence_threshold)

        def init_fn(position: PyTree, rng_key: PRNGKey):
            return cls.init(rng_key, position, logprob_fn, logprob_grad_fn)

        def step_fn(rng_key: PRNGKey, state):
            return step(
                rng_key, state, logprob_fn, step_size, alpha, delta, logprob_grad_fn
            )

        return SamplingAlgorithm(init_fn, step_fn)  # type: ignore[arg-type]


# -----------------------------------------------------------------------------
#                           VARIATIONAL INFERENCE
# -----------------------------------------------------------------------------


class pathfinder:
    """Implements the (basic) user interface for the pathfinder kernel.

    Pathfinder locates normal approximations to the target density along a
    quasi-Newton optimization path, with local covariance estimated using
    the inverse Hessian estimates produced by the L-BFGS optimizer.
    Pathfinder returns draws from the approximation with the lowest estimated
    Kullback-Leibler (KL) divergence to the true posterior.

    Note: all the heavy processing in performed in the init function, step
    function is just a drawing a sample from a normal distribution


    Returns
    -------
    A ``SamplingAlgorithm``.

    """

    init = staticmethod(vi.pathfinder.init)
    kernel = staticmethod(vi.pathfinder.kernel)

    def __new__(  # type: ignore[misc]
        cls,
        rng_key: PRNGKey,
        logprob_fn: Callable,
        num_samples: int = 200,
        **lbfgs_kwargs,
    ) -> SamplingAlgorithm:

        step = cls.kernel()

        def init_fn(position: PyTree):
            return cls.init(
                rng_key, logprob_fn, position, num_samples, False, **lbfgs_kwargs
            )

        def step_fn(rng_key: PRNGKey, state):
            return step(
                rng_key,
                state,
            )

        return SamplingAlgorithm(init_fn, step_fn)


def pathfinder_adaptation(
    algorithm: Union[hmc, nuts],
    logprob_fn: Callable,
    num_steps: int = 400,
    initial_step_size: float = 1.0,
    target_acceptance_rate: float = 0.65,
    **parameters,
) -> AdaptationAlgorithm:
    """Adapt the parameters of algorithms in the HMC family.

    Algorithms in the HMC family on a euclidean manifold depend on the value of
    at least two parameters: the step size, related to the trajectory
    integrator, and the mass matrix, linked to the euclidean metric.

    Good tuning is very important, especially for algorithms like NUTS which can
    be extremely inefficient with the wrong parameter values.
    This function tunes the values of these parameters according to this schema:
        * pathfinder algorithm is run and an estimation of the inverse mass matrix
          is derived, as well as an initialization point for the markov chain
        * Nesterov's dual averaging adaptation is then run to tune the step size

    Parameters
    ----------
    algorithm
        The algorithm whose parameters are being tuned.
    logprob_fn
        The log density probability density function from which we wish to sample.
    num_steps
        The number of adaptation steps for the dual averaging adaptation scheme.
    initial_step_size
        The initial step size used in the algorithm.
    target_acceptance_rate
        The acceptance rate that we target during step size adaptation.
    **parameters
        The extra parameters to pass to the algorithm, e.g. the number of
        integration steps for HMC.

    Returns
    -------
    A function that returns the last chain state and a sampling kernel with the tuned parameter values from an initial state.

    """

    kernel = algorithm.kernel()

    def kernel_factory(step_size: float, inverse_mass_matrix: Array):
        def kernel_fn(rng_key, state):
            return kernel(
                rng_key,
                state,
                logprob_fn,
                step_size,
                inverse_mass_matrix,
                **parameters,
            )

        return kernel_fn

    init, update, final = adaptation.pathfinder_adaptation.base(
        kernel_factory,
        logprob_fn,
        target_acceptance_rate=target_acceptance_rate,
    )

    @jax.jit
    def one_step(carry, rng_key):
        state, adaptation_state = carry
        state, adaptation_state, info = update(rng_key, state, adaptation_state)
        return ((state, adaptation_state), (state, info, adaptation_state.da_state))

    def run(rng_key: PRNGKey, position: PyTree):
        init_warmup_state, init_position = init(rng_key, position, initial_step_size)
        init_state = algorithm.init(init_position, logprob_fn)

        keys = jax.random.split(rng_key, num_steps)
        last_state, warmup_chain = jax.lax.scan(
            one_step,
            (init_state, init_warmup_state),
            keys,
        )
        last_chain_state, last_warmup_state = last_state
        history_state, history_info, history_da = warmup_chain
        history_adaptation = last_warmup_state._replace(da_state=history_da)

        warmup_chain = (history_state, history_info, history_adaptation)

        step_size, inverse_mass_matrix = final(last_warmup_state)
        kernel = kernel_factory(step_size, inverse_mass_matrix)

        return last_chain_state, kernel, warmup_chain

    return AdaptationAlgorithm(run)



"""Test the ess function"""
import functools
import itertools

import chex
import jax
import jax.numpy as jnp
import numpy as np
from absl.testing import absltest, parameterized
from jax.scipy.stats.multivariate_normal import logpdf as multivariate_logpdf
from jax.scipy.stats.norm import logpdf as univariate_logpdf

import blackjax.smc.ess as ess
import blackjax.smc.solver as solver


class SMCEffectiveSampleSizeTest(chex.TestCase):
    @chex.all_variants(with_pmap=False)
    @parameterized.parameters([100, 1000, 5000])
    def test_ess(self, N):
        log_ess_fn = self.variant(functools.partial(ess.ess, log=True))
        ess_fn = self.variant(functools.partial(ess.ess, log=False))

        w = np.random.rand(N)
        log_w = np.log(w)
        log_ess_val = log_ess_fn(log_w)
        ess_val = ess_fn(log_w)
        np.testing.assert_almost_equal(np.log(ess_val), log_ess_val, decimal=3)

        normalized_w = w / w.sum()
        log_normalized_w = np.log(normalized_w)
        log_normalized_ess_val = log_ess_fn(log_normalized_w)
        normalized_ess_val = ess_fn(log_normalized_w)

        np.testing.assert_almost_equal(log_ess_val, log_normalized_ess_val, decimal=3)
        np.testing.assert_almost_equal(ess_val, normalized_ess_val, decimal=3)
        np.testing.assert_almost_equal(
            ess_val, 1 / np.sum(normalized_w**2), decimal=3
        )

    @chex.all_variants(with_pmap=False)
    @parameterized.parameters(itertools.product([0.25, 0.5], [100, 1000, 5000]))
    def test_ess_solver(self, target_ess, N):
        potential_fn = lambda pytree: -univariate_logpdf(pytree, scale=0.1)
        potential = jax.vmap(lambda x: potential_fn(x), in_axes=[0])
        particles = np.random.normal(0, 1, size=(N, 1))
        self.ess_solver_test_case(potential, particles, target_ess, N, 1.0)

    @chex.all_variants(with_pmap=False)
    @parameterized.parameters(itertools.product([0.25, 0.5], [100, 1000, 5000]))
    def test_ess_solver_multivariate(self, target_ess, N):
        """
        Posterior with more than one variable. Let's assume we want to
        sample from P(x) x ~ N(mean, cov) x in R^{2}
        """
        mean = jnp.zeros((1, 2))
        cov = jnp.diag(jnp.array([1, 1]))
        potential_fn = lambda pytree: -multivariate_logpdf(pytree, mean=mean, cov=cov)
        potential = jax.vmap(lambda x: potential_fn(x), in_axes=[0], out_axes=0)
        particles = np.random.multivariate_normal(
            mean=[0.0, 0.0], cov=[[1.0, 0.0], [0.0, 1.0]], size=N
        )
        self.ess_solver_test_case(potential, particles, target_ess, N, 10.0)

    @chex.all_variants(with_pmap=False)
    @parameterized.parameters(itertools.product([0.25, 0.5], [100, 1000, 5000]))
    def test_ess_solver_posterior_signature(self, target_ess, N):
        """
        Posterior with more than one variable. Let's assume we want to
        sample from P(x,y) x ~ N(mean, cov) y ~ N(mean, cov)
        """
        mean = jnp.zeros((1, 2))
        cov = jnp.diag(jnp.array([1, 1]))

        def potential_fn(pytree):
            return -multivariate_logpdf(
                pytree[0], mean=mean, cov=cov
            ) - multivariate_logpdf(pytree[1], mean=mean, cov=cov)

        potential = jax.vmap(potential_fn, in_axes=[0], out_axes=0)
        particles = [
            np.random.multivariate_normal(
                mean=[0.0, 0.0], cov=[[1.0, 0.0], [0.0, 1.0]], size=N
            ),
            np.random.multivariate_normal(
                mean=[0.0, 0.0], cov=[[1.0, 0.0], [0.0, 1.0]], size=N
            ),
        ]
        self.ess_solver_test_case(potential, particles, target_ess, N, 10.0)

    def ess_solver_test_case(self, potential, particles, target_ess, N, max_delta):
        ess_solver_fn = functools.partial(
            ess.ess_solver,
            potential,
            target_ess=target_ess,
            max_delta=max_delta,
            root_solver=solver.dichotomy,
        )

        log_ess_solver = self.variant(
            functools.partial(ess_solver_fn, use_log_ess=True)
        )
        ess_solver = self.variant(functools.partial(ess_solver_fn, use_log_ess=False))
        delta_log = log_ess_solver(particles)
        delta = ess_solver(particles)
        assert delta_log > 0
        np.testing.assert_allclose(delta_log, delta, atol=1e-3, rtol=1e-3)
        log_ess = ess.ess(-delta_log * potential(particles), log=True)
        np.testing.assert_allclose(
            np.exp(log_ess), target_ess * N, atol=1e-1, rtol=1e-2
        )


if __name__ == "__main__":
    absltest.main()



"""Public API for marginal latent Gaussian sampling."""
from typing import NamedTuple

import jax
import jax.numpy as jnp
import jax.scipy.linalg as linalg

from blackjax.types import Array, PRNGKey

__all__ = ["MarginalState", "MarginalInfo", "init_and_kernel"]


class MarginalState(NamedTuple):
    """State of the RMH chain.

    x
        Current position of the chain.
    log_p_x
        Current value of the log-likelihood of the model
    grad_x
        Current value of the gradient of the log-likelihood of the model

    Other Attributes:
    -----------------
    U_x, U_grad_x: Array
        Auxiliary attributes

    """

    position: Array
    logprob: float
    logprob_grad: Array

    U_x: Array
    U_grad_x: Array


class MarginalInfo(NamedTuple):
    """Additional information on the RMH chain.

    This additional information can be used for debugging or computing
    diagnostics.

    acceptance_probability
        The acceptance probability of the transition, linked to the energy
        difference between the original and the proposed states.
    is_accepted
        Whether the proposed position was accepted or the original position
        was returned.
    proposal
        The state proposed by the proposal.
    """

    acceptance_probability: float
    is_accepted: bool
    proposal: MarginalState


def init_and_kernel(logprob_fn, covariance, mean=None):
    """Build the marginal version of the auxiliary gradient-based sampler

    Returns
    -------
    A kernel that takes a rng_key and a Pytree that contains the current state
    of the chain and that returns a new state of the chain along with
    information about the transition.
    An init function.

    """

    U, Gamma, U_t = jnp.linalg.svd(covariance, hermitian=True)

    if mean is not None:
        shift = linalg.solve(covariance, mean, sym_pos=True)
        val_and_grad = jax.value_and_grad(lambda x: logprob_fn(x) + jnp.dot(x, shift))
    else:
        val_and_grad = jax.value_and_grad(logprob_fn)

    def step(key: PRNGKey, state: MarginalState, delta):
        y_key, u_key = jax.random.split(key, 2)

        position, logprob, logprob_grad, U_x, U_grad_x = state

        # Update Gamma(delta)
        # TODO: Ideally, we could have a dichotomy, where we only update Gamma(delta) if delta changes,
        #       but this is hardly the most expensive part of the algorithm (the multiplication by U below is).
        Gamma_1 = Gamma * delta / (delta + 2 * Gamma)
        Gamma_3 = (delta + 2 * Gamma) / (delta + 4 * Gamma)
        Gamma_2 = Gamma_1 / Gamma_3

        # Propose a new y
        temp = Gamma_1 * (U_x / (0.5 * delta) + U_grad_x)
        temp = temp + jnp.sqrt(Gamma_2) * jax.random.normal(y_key, position.shape)
        y = U @ temp

        # Bookkeeping
        log_p_y, grad_y = val_and_grad(y)
        U_y = U_t @ y
        U_grad_y = U_t @ grad_y

        # Acceptance step
        temp_x = Gamma_1 * (U_x / (0.5 * delta) + 0.5 * U_grad_x)
        temp_y = Gamma_1 * (U_y / (0.5 * delta) + 0.5 * U_grad_y)

        hxy = jnp.dot(U_x - temp_y, Gamma_3 * U_grad_y)
        hyx = jnp.dot(U_y - temp_x, Gamma_3 * U_grad_x)

        alpha = jnp.minimum(1, jnp.exp(log_p_y - logprob + hxy - hyx))
        accept = jax.random.uniform(u_key) < alpha

        proposed_state = MarginalState(y, log_p_y, grad_y, U_y, U_grad_y)
        state = jax.lax.cond(accept, lambda _: proposed_state, lambda _: state, None)
        info = MarginalInfo(alpha, accept, proposed_state)
        return state, info

    def init(position):
        logprob, logprob_grad = val_and_grad(position)
        return MarginalState(
            position, logprob, logprob_grad, U_t @ position, U_t @ logprob_grad
        )

    return init, step



import numpy as np
import pytest

from blackjax.adaptation import window_adaptation


@pytest.mark.parametrize(
    "num_steps, expected_schedule",
    [
        (19, [(0, False)] * 19),  # no mass matrix adaptation
        (
            100,
            [(0, False)] * 15 + [(1, False)] * 74 + [(1, True)] + [(0, False)] * 10,
        ),  # windows are resized
        (
            200,
            [(0, False)] * 75
            + [(1, False)] * 24
            + [(1, True)]
            + [(1, False)] * 49
            + [(1, True)]
            + [(0, False)] * 50,
        ),
    ],
)
def test_adaptation_schedule(num_steps, expected_schedule):
    adaptation_schedule = window_adaptation.schedule(num_steps)
    assert num_steps == len(adaptation_schedule)
    assert np.array_equal(adaptation_schedule, expected_schedule)



"""Public API for the Generalized (Non-reversible w/ persistent momentum) HMC Kernel"""
from typing import Callable, NamedTuple, Optional, Tuple

import jax
import jax.numpy as jnp
from jax.flatten_util import ravel_pytree

import blackjax.mcmc.hmc as hmc
import blackjax.mcmc.integrators as integrators
import blackjax.mcmc.metrics as metrics
import blackjax.mcmc.proposal as proposal
from blackjax.types import PRNGKey, PyTree

__all__ = ["GHMCState", "init", "kernel"]


class GHMCState(NamedTuple):
    """State of the Generalized HMC algorithm.

    The Generalized HMC algorithm is persistent on its momentum, hence
    taking as input a position and momentum pair, updating and returning
    it for the next iteration. The algorithm also uses a persistent slice
    to perform a non-reversible Metropolis Hastings update, thus we also
    store the current slice variable and return its updated version after
    each iteration. To make computations more efficient, we also store
    the current potential energy as well as the current gradient of the
    potential energy.
    """

    position: PyTree
    momentum: PyTree
    potential_energy: float
    potential_energy_grad: PyTree
    slice: float


def init(
    rng_key: PRNGKey,
    position: PyTree,
    logprob_fn: Callable,
    logprob_grad_fn: Optional[Callable] = None,
):
    def potential_fn(x):
        return -logprob_fn(x)

    if logprob_grad_fn:
        potential_energy_grad = jax.tree_map(
            lambda g: -1.0 * g, logprob_grad_fn(position)
        )
        potential_energy = potential_fn(position)

    else:
        potential_energy, potential_energy_grad = jax.value_and_grad(potential_fn)(
            position
        )

    p, unravel_fn = ravel_pytree(position)
    key_mometum, key_slice = jax.random.split(rng_key)
    momentum = unravel_fn(jax.random.normal(key_mometum, p.shape))
    slice = jax.random.uniform(key_slice, minval=-1.0, maxval=1.0)

    return GHMCState(position, momentum, potential_energy, potential_energy_grad, slice)


def kernel(
    noise_gn: Callable = lambda _: 0.0,
    divergence_threshold: float = 1000,
):
    """Build a Generalized HMC kernel.

    The Generalized HMC kernel performs a similar procedure to the standard HMC
    kernel with the difference of a persistent momentum variable and a non-reversible
    Metropolis-Hastings step instead of the standard Metropolis-Hastings acceptance
    step. This means that; apart from momentum and slice variables that are dependent
    on the previous momentum and slice variables, and a Metropolis-Hastings step
    performed (equivalently) as slice sampling; the standard HMC's implementation can
    be re-used to perform Generalized HMC sampling.

    Parameters
    ----------
    noise_gn
        A function that takes as input the slice variable and outputs a random
        variable used as a noise correction of the persistent slice update.
        The parameter defaults to a random variable with a single atom at 0.
    divergence_threshold
        Value of the difference in energy above which we consider that the
        transition is divergent.

    Returns
    -------
    A kernel that takes a rng_key, a Pytree that contains the current state
    of the chain, and free parameters of the sampling mechanism; and that
    returns a new state of the chain along with information about the transition.
    """

    _, kinetic_energy_fn, _ = metrics.gaussian_euclidean(jnp.ones(1))
    sample_proposal = proposal.nonreversible_slice_sampling

    def one_step(
        rng_key: PRNGKey,
        state: GHMCState,
        logprob_fn: Callable,
        step_size: PyTree,  # float,
        alpha: float,
        delta: float,
        logprob_grad_fn: Optional[Callable] = None,
    ) -> Tuple[GHMCState, hmc.HMCInfo]:
        """Generate new sample with the Generalized HMC kernel.

        Parameters
        ----------
        rng_key
            JAX's pseudo random number generating key.
        state
            Current state of the chain.
        logprob_fn
            (Unnormalized) Log density function being targeted.
        step_size
            Pytree with the same structure as the targeted position variable
            specifying the step size used for each dimension of the target.
        alpha
            Variable specifying the degree of persistent momentum, complementary
            to independent new momentum.
        delta
            Fixed (non-random) amount of translation added at each new iteration
            to the slice variable for non-reversible slice sampling.
        logprob_grad_fn
            Optional function customizing the gradients of the target log density.
        """

        def potential_fn(x):
            return -logprob_fn(x)

        symplectic_integrator = velocity_verlet(
            potential_fn, kinetic_energy_fn, logprob_grad_fn
        )
        proposal_generator = hmc.hmc_proposal(
            symplectic_integrator,
            kinetic_energy_fn,
            step_size,
            divergence_threshold=divergence_threshold,
            sample_proposal=sample_proposal,
        )

        key_momentum, key_noise = jax.random.split(rng_key)
        position, momentum, potential_energy, potential_energy_grad, slice = state
        # New momentum is persistent
        momentum = update_momentum(key_momentum, state, alpha)
        # Slice is non-reversible
        slice = ((slice + 1.0 + delta + noise_gn(key_noise)) % 2) - 1.0

        integrator_state = integrators.IntegratorState(
            position, momentum, potential_energy, potential_energy_grad
        )
        proposal, info = proposal_generator(slice, integrator_state)
        proposal = hmc.flip_momentum(proposal)
        state = GHMCState(
            proposal.position,
            proposal.momentum,
            proposal.potential_energy,
            proposal.potential_energy_grad,
            info.acceptance_probability,
        )

        return state, info

    return one_step


def velocity_verlet(
    potential_fn: Callable,
    kinetic_energy_fn: integrators.EuclideanKineticEnergy,
    logprob_grad_fn: Optional[Callable] = None,
) -> integrators.EuclideanIntegrator:
    """The velocity Verlet (or Verlet-Störmer) integrator.

    Generates an implementation similar to the one in `integrators` except
    with a step_size parameter of the same PyTree structure as the position
    (target) variable that specifies an independent step_size for each dimension
    of the target.
    """

    if logprob_grad_fn:
        potential_and_grad_fn = lambda x: (
            potential_fn(x),
            jax.tree_map(lambda g: -1.0 * g, logprob_grad_fn(x)),
        )
    else:
        potential_and_grad_fn = jax.value_and_grad(potential_fn)
    kinetic_energy_grad_fn = jax.grad(kinetic_energy_fn)

    def one_step(
        state: integrators.IntegratorState, step_size: PyTree
    ) -> integrators.IntegratorState:
        position, momentum, _, potential_energy_grad = state

        momentum = jax.tree_util.tree_map(
            lambda momentum, potential_grad, step_size: momentum
            - 0.5 * step_size * potential_grad,
            momentum,
            potential_energy_grad,
            step_size,
        )

        kinetic_grad = kinetic_energy_grad_fn(momentum)
        position = jax.tree_util.tree_map(
            lambda position, kinetic_grad, step_size: position
            + step_size * kinetic_grad,
            position,
            kinetic_grad,
            step_size,
        )

        potential_energy, potential_energy_grad = potential_and_grad_fn(position)
        momentum = jax.tree_util.tree_map(
            lambda momentum, potential_grad, step_size: momentum
            - 0.5 * step_size * potential_grad,
            momentum,
            potential_energy_grad,
            step_size,
        )

        return integrators.IntegratorState(
            position,
            momentum,
            potential_energy,
            potential_energy_grad,
        )

    return one_step


def update_momentum(rng_key, state, alpha):
    """Persistent update of the momentum variable.

    Performs a persistent update of the momentum, taking as input the previous
    momentum, a random number generating key and the parameter alpha. Outputs
    an updated momentum that is a mixture of the previous momentum a new sample
    from a Gaussian density (dependent on alpha). The weights of the mixture of
    these two components are a function of alpha.
    """

    position, momentum, *_ = state

    m, _ = ravel_pytree(momentum)
    momentum_generator, *_ = metrics.gaussian_euclidean(
        1 / alpha * jnp.ones(jnp.shape(m))
    )
    momentum = jax.tree_map(
        lambda prev_momentum, shifted_momentum: prev_momentum * jnp.sqrt(1.0 - alpha)
        + shifted_momentum,
        momentum,
        momentum_generator(rng_key, position),
    )

    return momentum



"""Test the accuracy of the MCMC kernels."""
import functools
import itertools

import chex
import jax
import jax.numpy as jnp
import jax.scipy.stats as stats
import numpy as np
from absl.testing import absltest, parameterized

import blackjax
import blackjax.diagnostics as diagnostics


def inference_loop(kernel, num_samples, rng_key, initial_state):
    def one_step(state, rng_key):
        state, _ = kernel(rng_key, state)
        return state, state

    keys = jax.random.split(rng_key, num_samples)
    _, states = jax.lax.scan(one_step, initial_state, keys)

    return states


def orbit_samples(orbits, weights, rng_key):
    def sample_orbit(orbit, weights, rng_key):
        sample = jax.random.choice(rng_key, orbit, p=weights)
        return sample

    keys = jax.random.split(rng_key, orbits.shape[0])
    samples = jax.vmap(sample_orbit)(orbits, weights, keys)

    return samples


def irmh_proposal_distribution(rng_key):
    """
    The proposal distribution is chosen to be wider than the target, so that the RMH rejection
    doesn't make the sample overemphasize the center of the target distribution.
    """
    return 1.0 + jax.random.normal(rng_key) * 25.0


regression_test_cases = [
    {
        "algorithm": blackjax.hmc,
        "initial_position": {"scale": 1.0, "coefs": 2.0},
        "parameters": {"num_integration_steps": 90},
        "num_warmup_steps": 1_000,
        "num_sampling_steps": 3_000,
    },
    {
        "algorithm": blackjax.nuts,
        "initial_position": {"scale": 1.0, "coefs": 2.0},
        "parameters": {},
        "num_warmup_steps": 1_000,
        "num_sampling_steps": 1_000,
    },
]


class LinearRegressionTest(chex.TestCase):
    """Test sampling of a linear regression model."""

    def setUp(self):
        super().setUp()
        self.key = jax.random.PRNGKey(19)

    def regression_logprob(self, scale, coefs, preds, x):
        """Linear regression"""
        scale_prior = stats.expon.logpdf(scale, 1, 1)
        coefs_prior = stats.norm.logpdf(coefs, 0, 5)
        y = jnp.dot(x, coefs)
        logpdf = stats.norm.logpdf(preds, y, scale)
        # reduce sum otherwise broacasting will make the logprob biased.
        return sum(x.sum() for x in [scale_prior, coefs_prior, logpdf])

    @parameterized.parameters(itertools.product(regression_test_cases, [True, False]))
    def test_window_adaptation(self, case, is_mass_matrix_diagonal):
        """Test the HMC kernel and the Stan warmup."""
        rng_key, init_key0, init_key1 = jax.random.split(self.key, 3)
        x_data = jax.random.normal(init_key0, shape=(1000, 1))
        y_data = 3 * x_data + jax.random.normal(init_key1, shape=x_data.shape)

        logposterior_fn_ = functools.partial(
            self.regression_logprob, x=x_data, preds=y_data
        )
        logposterior_fn = lambda x: logposterior_fn_(**x)

        warmup_key, inference_key = jax.random.split(rng_key, 2)

        warmup = blackjax.window_adaptation(
            case["algorithm"],
            logposterior_fn,
            case["num_warmup_steps"],
            is_mass_matrix_diagonal,
            progress_bar=True,
            **case["parameters"],
        )
        state, kernel, _ = warmup.run(
            warmup_key,
            case["initial_position"],
        )

        states = inference_loop(
            kernel, case["num_sampling_steps"], inference_key, state
        )

        coefs_samples = states.position["coefs"]
        scale_samples = states.position["scale"]

        np.testing.assert_allclose(np.mean(scale_samples), 1.0, atol=1e-1)
        np.testing.assert_allclose(np.mean(coefs_samples), 3.0, atol=1e-1)

    def test_mala(self):
        """Test the MALA kernel."""
        rng_key, init_key0, init_key1 = jax.random.split(self.key, 3)
        x_data = jax.random.normal(init_key0, shape=(1000, 1))
        y_data = 3 * x_data + jax.random.normal(init_key1, shape=x_data.shape)

        logposterior_fn_ = functools.partial(
            self.regression_logprob, x=x_data, preds=y_data
        )
        logposterior_fn = lambda x: logposterior_fn_(**x)

        warmup_key, inference_key = jax.random.split(rng_key, 2)

        mala = blackjax.mala(logposterior_fn, 1e-5)
        state = mala.init({"coefs": 1.0, "scale": 2.0})
        states = inference_loop(mala.step, 10_000, inference_key, state)

        coefs_samples = states.position["coefs"][3000:]
        scale_samples = states.position["scale"][3000:]

        np.testing.assert_allclose(np.mean(scale_samples), 1.0, atol=1e-1)
        np.testing.assert_allclose(np.mean(coefs_samples), 3.0, atol=1e-1)

    @parameterized.parameters(regression_test_cases)
    def test_pathfinder_adaptation(
        self,
        algorithm,
        num_warmup_steps,
        initial_position,
        num_sampling_steps,
        parameters,
    ):
        """Test the HMC kernel and the Stan warmup."""
        rng_key, init_key0, init_key1 = jax.random.split(self.key, 3)
        x_data = jax.random.normal(init_key0, shape=(1000, 1))
        y_data = 3 * x_data + jax.random.normal(init_key1, shape=x_data.shape)

        logposterior_fn_ = functools.partial(
            self.regression_logprob, x=x_data, preds=y_data
        )
        logposterior_fn = lambda x: logposterior_fn_(**x)

        warmup_key, inference_key = jax.random.split(rng_key, 2)

        warmup = blackjax.pathfinder_adaptation(
            algorithm,
            logposterior_fn,
            num_warmup_steps,
            **parameters,
        )
        state, kernel, _ = warmup.run(
            warmup_key,
            initial_position,
        )

        states = inference_loop(kernel, num_sampling_steps, inference_key, state)

        coefs_samples = states.position["coefs"]
        scale_samples = states.position["scale"]

        np.testing.assert_allclose(np.mean(scale_samples), 1.0, atol=1e-1)
        np.testing.assert_allclose(np.mean(coefs_samples), 3.0, atol=1e-1)

    def test_meads(self):
        """Test the MEADS adaptation w/ GHMC kernel."""
        rng_key, init_key0, init_key1 = jax.random.split(self.key, 3)
        x_data = jax.random.normal(init_key0, shape=(1000, 1))
        y_data = 3 * x_data + jax.random.normal(init_key1, shape=x_data.shape)

        logposterior_fn_ = functools.partial(
            self.regression_logprob, x=x_data, preds=y_data
        )
        logposterior_fn = lambda x: logposterior_fn_(**x)

        init_key, warmup_key, inference_key = jax.random.split(rng_key, 3)

        warmup = blackjax.meads(
            logposterior_fn,
            num_chain=128,
            num_steps=100,
        )
        scale_key, coefs_key = jax.random.split(init_key, 2)
        scales = 1.0 + jax.random.normal(scale_key, (128,))
        coefs = 3.0 + jax.random.normal(coefs_key, (128,))
        initial_positions = {"scale": scales, "coefs": coefs}
        chains_state, kernel, _ = warmup.run(
            warmup_key,
            initial_positions,
        )

        states = jax.vmap(
            lambda state: inference_loop(kernel, 100, inference_key, state)
        )(chains_state.states)

        coefs_samples = states.position["coefs"]
        scale_samples = states.position["scale"]

        np.testing.assert_allclose(np.mean(scale_samples), 1.0, atol=1e-1)
        np.testing.assert_allclose(np.mean(coefs_samples), 3.0, atol=1e-1)


class SGMCMCTest(chex.TestCase):
    """Test sampling of a linear regression model."""

    def setUp(self):
        super().setUp()
        self.key = jax.random.PRNGKey(19)

    def logprior_fn(self, position):
        return -0.5 * jnp.dot(position, position) * 0.01

    def loglikelihood_fn(self, position, x):
        w = x - position
        return -0.5 * jnp.dot(w, w)

    def constant_step_size(_):
        return 1e-3

    @parameterized.parameters((1e-3, False), (constant_step_size, False), (1, True))
    def test_linear_regression_sgld(self, learning_rate, error_expected):
        """Test the HMC kernel and the Stan warmup."""
        import blackjax.sgmcmc.gradients

        rng_key, data_key = jax.random.split(self.key, 2)

        data_size = 1000
        X_data = jax.random.normal(data_key, shape=(data_size, 5))

        grad_fn = blackjax.sgmcmc.gradients.grad_estimator(
            self.logprior_fn, self.loglikelihood_fn, data_size
        )

        if error_expected:
            self.assertRaises(TypeError, blackjax.sgld(grad_fn, learning_rate))
            return

        sgld = blackjax.sgld(grad_fn, learning_rate)
        init_position = 1.0
        data_batch = X_data[:100, :]
        init_state = sgld.init(init_position, data_batch)

        _, rng_key = jax.random.split(rng_key)
        data_batch = X_data[100:200, :]
        _ = sgld.step(rng_key, init_state, data_batch)

    @parameterized.parameters((1e-3, False), (constant_step_size, False), (1, True))
    def test_linear_regression_sghmc(self, learning_rate, error_expected):
        """Test the HMC kernel and the Stan warmup."""
        import blackjax.sgmcmc.gradients

        rng_key, data_key = jax.random.split(self.key, 2)

        data_size = 1000
        X_data = jax.random.normal(data_key, shape=(data_size, 5))

        grad_fn = blackjax.sgmcmc.gradients.grad_estimator(
            self.logprior_fn, self.loglikelihood_fn, data_size
        )

        if error_expected:
            self.assertRaises(TypeError, blackjax.sgld(grad_fn, learning_rate))
            return

        sghmc = blackjax.sghmc(grad_fn, learning_rate, 10)

        init_position = 1.0
        data_batch = X_data[:100, :]
        init_state = sghmc.init(init_position, data_batch)

        _, rng_key = jax.random.split(rng_key)
        data_batch = X_data[100:200, :]
        _ = sghmc.step(rng_key, init_state, data_batch)


class LatentGaussianTest(chex.TestCase):
    """Test sampling of a linear regression model."""

    def setUp(self):
        super().setUp()
        self.key = jax.random.PRNGKey(19)
        self.C = 2.0 * np.eye(1)
        self.delta = 5.0
        self.sampling_steps = 25_000
        self.burnin = 5_000

    @chex.all_variants(with_pmap=False)
    def test_latent_gaussian(self):
        from blackjax import mgrad_gaussian

        init, step = mgrad_gaussian(lambda x: -0.5 * jnp.sum((x - 1.0) ** 2), self.C)

        kernel = lambda key, x: step(key, x, self.delta)
        initial_state = init(jnp.zeros((1,)))

        states = self.variant(
            functools.partial(inference_loop, kernel, self.sampling_steps),
        )(self.key, initial_state)

        np.testing.assert_allclose(
            np.var(states.position[self.burnin :]), 1 / (1 + 0.5), rtol=1e-2, atol=1e-2
        )
        np.testing.assert_allclose(
            np.mean(states.position[self.burnin :]), 2 / 3, rtol=1e-2, atol=1e-2
        )


normal_test_cases = [
    {
        "algorithm": blackjax.hmc,
        "initial_position": jnp.array(3.0),
        "parameters": {
            "step_size": 3.9,
            "inverse_mass_matrix": jnp.array([1.0]),
            "num_integration_steps": 30,
        },
        "num_sampling_steps": 6000,
        "burnin": 1_000,
    },
    {
        "algorithm": blackjax.nuts,
        "initial_position": jnp.array(3.0),
        "parameters": {"step_size": 4.0, "inverse_mass_matrix": jnp.array([1.0])},
        "num_sampling_steps": 6000,
        "burnin": 1_000,
    },
    {
        "algorithm": blackjax.orbital_hmc,
        "initial_position": jnp.array(100.0),
        "parameters": {
            "step_size": 0.1,
            "inverse_mass_matrix": jnp.array([0.1]),
            "period": 100,
        },
        "num_sampling_steps": 20_000,
        "burnin": 15_000,
    },
    {
        "algorithm": blackjax.rmh,
        "initial_position": 1.0,
        "parameters": {"sigma": jnp.array([1.0])},
        "num_sampling_steps": 20_000,
        "burnin": 5_000,
    },
    {
        "algorithm": blackjax.mala,
        "initial_position": 1.0,
        "parameters": {"step_size": 1e-1},
        "num_sampling_steps": 20_000,
        "burnin": 2_000,
    },
    {
        "algorithm": blackjax.elliptical_slice,
        "initial_position": 1.0,
        "parameters": {"cov": jnp.array([2.0**2]), "mean": 1.0},
        "num_sampling_steps": 20_000,
        "burnin": 5_000,
    },
    {
        "algorithm": blackjax.irmh,
        "initial_position": jnp.array(1.0),
        "parameters": {},
        "num_sampling_steps": 50_000,
        "burnin": 5_000,
    },
    {
        "algorithm": blackjax.ghmc,
        "initial_position": jnp.array(1.0),
        "parameters": {
            "step_size": jnp.array(1.0),
            "alpha": 0.8,
            "delta": 2.0,
        },
        "num_sampling_steps": 6000,
        "burnin": 1_000,
    },
]


class UnivariateNormalTest(chex.TestCase):
    """Test sampling of a univariate Normal distribution."""

    def setUp(self):
        super().setUp()
        self.key = jax.random.PRNGKey(19)

    def normal_logprob(self, x):
        return stats.norm.logpdf(x, loc=1.0, scale=2.0)

    @chex.all_variants(with_pmap=False)
    @parameterized.parameters(normal_test_cases)
    def test_univariate_normal(
        self, algorithm, initial_position, parameters, num_sampling_steps, burnin
    ):
        if algorithm == blackjax.irmh:
            parameters["proposal_distribution"] = irmh_proposal_distribution

        algo = algorithm(self.normal_logprob, **parameters)
        if algorithm == blackjax.elliptical_slice:
            algo = algorithm(lambda _: 1.0, **parameters)
        if algorithm == blackjax.ghmc:
            initial_state = algo.init(initial_position, self.key)
        else:
            initial_state = algo.init(initial_position)

        kernel = algo.step
        states = self.variant(
            functools.partial(inference_loop, kernel, num_sampling_steps)
        )(self.key, initial_state)

        if algorithm == blackjax.orbital_hmc:
            _, orbit_key = jax.random.split(self.key)
            samples = orbit_samples(
                states.positions[burnin:], states.weights[burnin:], orbit_key
            )
        else:
            samples = states.position[burnin:]

        np.testing.assert_allclose(np.mean(samples), 1.0, rtol=1e-1)
        np.testing.assert_allclose(np.var(samples), 4.0, rtol=1e-1)


mcse_test_cases = [
    {
        "algorithm": blackjax.hmc,
        "parameters": {
            "step_size": 1.0,
            "num_integration_steps": 32,
        },
        "custom_gradients": False,
    },
    {
        "algorithm": blackjax.nuts,
        "parameters": {"step_size": 1.0},
        "custom_gradients": False,
    },
    {
        "algorithm": blackjax.nuts,
        "parameters": {"step_size": 1.0},
        "custom_gradients": True,
    },
]


class MonteCarloStandardErrorTest(chex.TestCase):
    """Test sampler correctness using Monte Carlo Central Limit Theorem."""

    def setUp(self):
        super().setUp()
        self.key = jax.random.PRNGKey(20220203)

    def generate_multivariate_target(self, rng=None):
        """Genrate a Multivariate Normal distribution as target."""
        if rng is None:
            loc = jnp.array([0.0, 3])
            scale = jnp.array([1.0, 2.0])
            rho = jnp.array(0.75)
        else:
            rng, loc_rng, scale_rng, rho_rng = jax.random.split(rng, 4)
            loc = jax.random.normal(loc_rng, [2]) * 10.0
            scale = jnp.abs(jax.random.normal(scale_rng, [2])) * 2.5
            rho = jax.random.uniform(rho_rng, [], minval=-1.0, maxval=1.0)

        cov = jnp.diag(scale**2)
        cov = cov.at[0, 1].set(rho * scale[0] * scale[1])
        cov = cov.at[1, 0].set(rho * scale[0] * scale[1])

        def logprob_fn(x):
            return stats.multivariate_normal.logpdf(x, loc, cov).sum()

        return logprob_fn, loc, scale, rho

    def mcse_test(self, samples, true_param, p_val=0.01):
        posterior_mean = jnp.mean(samples, axis=[0, 1])
        ess = diagnostics.effective_sample_size(samples, chain_axis=0, sample_axis=1)
        posterior_sd = jnp.std(samples, axis=0, ddof=1)
        avg_monte_carlo_standard_error = jnp.mean(posterior_sd, axis=0) / jnp.sqrt(ess)
        scaled_error = (
            jnp.abs(posterior_mean - true_param) / avg_monte_carlo_standard_error
        )
        np.testing.assert_array_less(scaled_error, stats.norm.ppf(1 - p_val))
        return scaled_error

    @parameterized.parameters(mcse_test_cases)
    def test_mcse(self, algorithm, parameters, custom_gradients):
        """Test convergence using Monte Carlo CLT across multiple chains."""
        init_fn_key, pos_init_key, sample_key = jax.random.split(self.key, 3)
        logprob_fn, true_loc, true_scale, true_rho = self.generate_multivariate_target(
            None
        )
        if custom_gradients:
            logprob_grad_fn = jax.jacfwd(logprob_fn)
        else:
            logprob_grad_fn = None

        kernel = algorithm(
            logprob_fn,
            inverse_mass_matrix=true_scale,
            logprob_grad_fn=logprob_grad_fn,
            **parameters,
        )

        num_chains = 10
        initial_positions = jax.random.normal(pos_init_key, [num_chains, 2])
        initial_states = jax.vmap(kernel.init, in_axes=(0,))(initial_positions)
        multi_chain_sample_key = jax.random.split(sample_key, num_chains)

        inference_loop_multiple_chains = jax.vmap(
            functools.partial(inference_loop, kernel.step, 2_000)
        )
        states = inference_loop_multiple_chains(multi_chain_sample_key, initial_states)

        posterior_samples = states.position[:, -1000:]
        posterior_delta = posterior_samples - true_loc
        posterior_variance = posterior_delta**2.0
        posterior_correlation = jnp.prod(posterior_delta, axis=-1, keepdims=True) / (
            true_scale[0] * true_scale[1]
        )

        _ = jax.tree_map(
            self.mcse_test,
            [posterior_samples, posterior_variance, posterior_correlation],
            [true_loc, true_scale**2, true_rho],
        )


if __name__ == "__main__":
    absltest.main()



"""Public API for the Independent Rosenbluth-Metropolis-Hastings kernels."""
from typing import Callable, Tuple

from blackjax.mcmc.rmh import RMHInfo, RMHState, rmh
from blackjax.types import PRNGKey, PyTree

__all__ = ["kernel"]


def kernel(proposal_distribution: Callable) -> Callable:
    """
    Build an Independent Random Walk Rosenbluth-Metropolis-Hastings kernel. This implies
    that the proposal distribution does not depend on the particle being mutated.
    Reference: Algorithm 2 from https://arxiv.org/pdf/2008.02455.pdf

    Parameters
    ----------
    proposal_distribution
        A function that, given a PRNGKey, is able to produce a sample in the same
        domain of the target distribution.

    Returns
    -------
    A kernel that takes a rng_key and a Pytree that contains the current state
    of the chain and that returns a new state of the chain along with
    information about the transition.

    """

    def one_step(
        rng_key: PRNGKey, state: RMHState, logprob_fn: Callable
    ) -> Tuple[RMHState, RMHInfo]:
        def proposal_generator(rng_key: PRNGKey, position: PyTree):
            return proposal_distribution(rng_key)

        kernel = rmh(logprob_fn, proposal_generator)
        return kernel(rng_key, state)

    return one_step



"""Public API for the Stochastic gradient Hamiltonian Monte Carlo kernel."""
from typing import Callable

import jax
import jax.numpy as jnp

from blackjax.sgmcmc.diffusion import SGHMCState, sghmc
from blackjax.sgmcmc.sgld import SGLDState
from blackjax.types import PRNGKey, PyTree

__all__ = ["kernel"]


def sample_momentum(rng_key: PRNGKey, position: PyTree, step_size: float):
    position_flat, unravel_fn = jax.flatten_util.ravel_pytree(position)
    noise_flat = jnp.sqrt(step_size) * jax.random.normal(
        rng_key, shape=jnp.shape(position_flat)
    )
    return unravel_fn(noise_flat)


def kernel(
    grad_estimator_fn: Callable, alpha: float = 0.01, beta: float = 0
) -> Callable:
    integrator = sghmc(grad_estimator_fn)

    def one_step(
        rng_key: PRNGKey, state: SGLDState, data_batch: PyTree, step_size: float, L: int
    ) -> SGLDState:

        step, position, logprob_grad = state
        momentum = sample_momentum(rng_key, position, step_size)
        diffusion_state = SGHMCState(position, momentum, logprob_grad)

        def body_fn(state, rng_key):
            new_state = integrator(rng_key, state, step_size, data_batch)
            return new_state, new_state

        keys = jax.random.split(rng_key, L)
        last_state, _ = jax.lax.scan(body_fn, diffusion_state, keys)

        return SGLDState(step + 1, last_state.position, last_state.logprob_grad)

    return one_step
